/*
 * moltiblock
 *
 * By Thomas Habets <thomas@habets.pp.se>
 *
 * User-space plugin to netfilter for blocking lots of nets.
 * No need to recompile kernel or patch other sources.
 *
 * $Id: moltiblock.cc 1180 2004-12-13 22:31:26Z marvin $
 */
/*
 *  Copyright (C) 2004 Thomas Habets <thomas@habets.pp.se>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
extern "C" {
#include <linux/netfilter.h>
#include <libipq.h>
}

#include <stdio.h>
#include <time.h>
#include <netinet/in.h>

#include <sys/time.h>

#include <vector>
#include <algorithm>
#include <map>

using namespace std;

static const double version = 0.10f;

typedef pair<u_int32_t, int> denyentry_t;
static vector<denyentry_t> deny;
static map<u_int32_t,u_int64_t> stats;
static unsigned int verbose = 0;

static const char *def_denyfile = "/etc/moltiblock/cidr.deny";
//static const char *def_statsfile = "/var/log/moltiblock.stats";
static const char *def_statsfile = NULL;
static const char *statsfile = def_statsfile;

/*
 *
 */
static void die(struct ipq_handle *h)
{
        ipq_perror("passer");
        ipq_destroy_handle(h);
        exit(1);
}

/*
 *
 */
static bool lessthan(const denyentry_t &a, const denyentry_t &b)
{
	u_int32_t m = a.second & b.second;

	if (verbose) {
		printf("test: %.8x/%.8x < %.8x/%.8x\n",
		       a.first, a.second, b.first,b.second);
	}
	if ((a.first & m) == (b.first & m)) {
		if (verbose) {	printf("... false\n");	}
		return false;
	}
	if (a.first < b.first) {
		if (verbose) {	printf("... true\n");	}
		return true;
	}
	if (verbose) {	printf("... false\n");	}
	return false;
}

/*
 *
 */
static void init_deny(const char *filename)
{
	FILE *f;
	char buf[1024];
	if (!(f = fopen(filename, "r"))) {
		fprintf(stderr, "fopen(%s)\n", filename);
		exit(1);
	}
	while (fgets(buf, sizeof(buf)-1, f)) {
		char *s;
		u_int32_t mask = 0xffffffff;
		union {
			u_int32_t ip;
			struct { char a,b,c,d;	};
		} foo;
		if (4 != sscanf(buf, "%d.%d.%d.%d",
				&foo.a,&foo.b,&foo.c,&foo.d)) {
			continue;
		}
		foo.ip = htonl(foo.ip);
		if ((s = strchr(buf, '/'))) {
			mask = ~((1 << (32-atoi(s+1))) - 1);
		}
		deny.push_back(denyentry_t(foo.ip,mask));
	}
	sort(deny.begin(), deny.end(), lessthan);
}

/*
 *
 */
static bool test_deny(u_int32_t val)
{
	denyentry_t v;
	v.first = val;
	v.second = 0xffffffff;
	if (binary_search(deny.begin(), deny.end(), v, lessthan)) {
		if (statsfile) {
			stats[val]++;
		}
		if (verbose) {	printf("Blocked %.8x\n", val);	}
		return true;
	}
	return false;
}

/*
 *
 */
static void dumpstats(const char *filename)
{
	if (!statsfile) {
		return;
	}

	FILE *f;
	union {
		u_int32_t ip;
		struct { char a,b,c,d;	};
	} foo;

	if (!(f = fopen(filename, "w"))) {
		return;
	}
	for (map<u_int32_t,u_int64_t>::const_iterator itr = stats.begin();
	     itr != stats.end();
	     ++itr) {
		foo.ip = itr->first;
		fprintf(f, "%hhu.%hhu.%hhu.%hhu %llu\n",
			foo.d,foo.c,foo.b,foo.a,itr->second);
	}
	fclose(f);
}

/*
 *
 */
static unsigned int verdict(const unsigned char *buf)
{
	ipq_packet_msg_t *p;
	p = ipq_get_packet(buf);
	if (p->data_len < 16) {
		return NF_ACCEPT;
	}

	u_int32_t src,dst;
	src = htonl(((u_int32_t*)p->payload)[3]);
	dst = htonl(((u_int32_t*)p->payload)[4]);
	
	if (test_deny(src) || test_deny(dst)) {
		if (verbose) { printf("dropped: %p -> %p\n", src,dst); }
		return NF_DROP;
	}

	return NF_ACCEPT;
}

/*
 *
 */
static void usage(const char *av0, int err)
{
	printf("Moltiblock %.2f, by Thomas Habets <thomas@habets.pp.se>\n"
	       "Usage: %s [ -Dhv ] [ -f <denyfile> ] [ -s <statsfile ]\n"
	       "\t-D                Run in foreground\n"
	       "\t-f <file>         Deny-file (default: %s)\n"
	       "\t-h                This help message\n"
	       "\t-s <statsfile>    Statsfile (default: %s)\n"
	       "\t-v                Increase verbosity factor\n",
	       version, av0, def_denyfile,
	       def_statsfile?def_statsfile : "no stats");
	exit(err);
}


/*
 *
 */
int main(int argc, char **argv)
{
        int status;
        unsigned char buf[2048];
        struct ipq_handle *h;
	int packetcounter = 0;
	int c;
	const char *denyfile = def_denyfile;
	bool daemon = true;

	while (EOF != (c = getopt(argc, argv, "Df:hs:v"))) {
		switch(c) {
		case 'D':
			daemon = false;
			break;
		case 'f':
			denyfile = optarg;
			break;
		case 'h':
			usage(argv[0], 0);
			break;
		case 's':
			statsfile = optarg;
			break;
		case 'v':
			verbose++;
			break;
		default:
			usage(argv[0], 1);
		}
	}

	init_deny(denyfile);
	if (verbose) { printf("Loaded blocklist of size %d\n", deny.size());}

	if (verbose) {
		printf("yes: %d\n", test_deny(0xD8CF5040)); // 216.207.80.64
		printf("yes: %d\n", test_deny(0x02000001));
		printf("no: %d\n", test_deny(0x0A000001));
	}
        
        if (!(h = ipq_create_handle(0, PF_INET))) {
                die(h);
	}
                
        if (0 > (status = ipq_set_mode(h, IPQ_COPY_PACKET, sizeof(buf)))) {
                die(h);
	}

	if (daemon && fork()) {
		exit(0);
	}
                
	struct timeval last_dump = {0,0};
        for(u_int64_t packetcounter = 1; ; packetcounter++) {
                if (0 > (status = ipq_read(h, buf, sizeof(buf), 1000000))) {
                        die(h);
		}
		if (status) {
			switch (ipq_message_type(buf)) {
                        case NLMSG_ERROR:
                                fprintf(stderr, "Received error message %d\n",
                                        ipq_get_msgerr(buf));
                                break;
                        case IPQM_PACKET: {
                                ipq_packet_msg_t *m = ipq_get_packet(buf);
                                if (0 > (status = ipq_set_verdict(h,
								  m->packet_id,
								  verdict(buf),
								  0, NULL))) {
                                        die(h);
				}
                                break;
                        }
                        default:
                                fprintf(stderr, "Unknown message type!\n");
                                break;
			}
		}

		struct timeval tv;
		if (-1 == gettimeofday(&tv, NULL)) {
			fprintf(stderr, "gettimeofday() failed: %m\n");
			die(h);
		}
		if (tv.tv_sec - last_dump.tv_sec > 60) {
			if (-1 == gettimeofday(&last_dump, NULL)) {
				fprintf(stderr, "gettimeofday() failed: %m\n");
				die(h);
			}
			dumpstats(statsfile);
		}
        }
        ipq_destroy_handle(h);
        return 0;
}
